/* adal_switch.c */

#include <Pilot.h>

#include "adal.h"
#include "adal_rcp.h"


/*
 * constants
 */

#define FTR_ORIG TRAP2_ID

#define APP_DB_TYPE 'appl'
#define LAUNCHER_CREATOR 'lnch'

#define LAUNCHPAD_CREATOR 'LcPd'	/* LaunchPad */
#define COLAUNCHER_CREATOR 'CoLn'	/* CoLauncher */
#define CLICK_LC_CREATOR 'CLLC'		/* Click LC */
#define CLICK_LC_A_CREATOR 'CLCa'	/* Click LC Helper A */
#define CLICK_LC_B_CREATOR 'CLCb'	/* Click LC Helper B */
#define CLICK_LC_C_CREATOR 'CLCc'	/* Click LC Helper C */
#define CLICK_LC_D_CREATOR 'CLCd'	/* Click LC Helper D */

#define DEFAULT_CARD_NO 0	/* zzz */
#define HISTORY_DB_NAME "App/DA Launcher History"

#define HISTORY_RSRC_TYPE 'HIST'
#define HISTORY_RSRC_ID 1000
#define HISTORY_RSRC_SIZE (sizeof(struct history_record) \
	    + sizeof(struct history_entry) * HISTORY_MAX_COUNT)


/*
 * functions
 */

static void get_icon_name(UInt card_no, LocalID db_id, char *name)
{
    DmOpenRef db;
    VoidHand h;
    VoidPtr p;

    db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);
    if (db != NULL) {
	h = DmGet1Resource(ainRsc, ainID);
	if (h != NULL) {
	    p = MemHandleLock(h);
	    if (p != NULL) {
		StrNCopy(name, p, dmDBNameLength - 1);
		name[dmDBNameLength - 1] = '\0';
		MemHandleUnlock(h);
	    }
	    DmReleaseResource(h);
	}
	DmCloseDatabase(db);
    }
}

static void record_history(ULong type, ULong creator, char *name)
{
    DmOpenRef db;
    Boolean res_db;
    VoidHand h;
    struct history_record *p;
    struct history_record record;
    struct history_entry entry;
    int i;

    entry.type = type;
    entry.creator = creator;
    MemSet(entry.name, sizeof entry.name, 0);
    StrCopy(entry.name, name);

    db = DmOpenDatabaseByTypeCreator(HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
    if (db == NULL) {
	DWord rom_version;
	UInt card_no;
	LocalID db_id;
	ULong cur_type, cur_creator;

	if (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) == noErr
	 && rom_version >= sysMakeROMVersion(3, 5, 0, 0, 0)
	 && SysCurAppDatabase(&card_no, &db_id) == noErr
	 && DmDatabaseInfo(card_no, db_id, NULL, NULL, NULL, NULL,
		NULL, NULL, NULL, NULL, NULL, &cur_type, &cur_creator) == noErr
	 && cur_type == APP_DB_TYPE
	 && cur_creator == LAUNCHER_CREATOR) {
	    return;	/* to avoid crash on Palm OS 3.5 launcher... */
	}

	if (DmCreateDatabase(DEFAULT_CARD_NO, HISTORY_DB_NAME,
		CREATOR, HISTORY_DB_TYPE, true) == noErr) {
	    db = DmOpenDatabaseByTypeCreator(
		    HISTORY_DB_TYPE, CREATOR, dmModeReadWrite);
	}
    }
    if (db != NULL) {
	if (DmOpenDatabaseInfo(db, NULL, NULL, NULL, NULL, &res_db) == noErr
	 && res_db != false) {
	    h = DmGet1Resource(HISTORY_RSRC_TYPE, HISTORY_RSRC_ID);
	    if (h == NULL) {
		h = DmNewResource(db,
			HISTORY_RSRC_TYPE, HISTORY_RSRC_ID, HISTORY_RSRC_SIZE);
		if (h != NULL
		 && (p = MemHandleLock(h)) != NULL) {
		    DmSet(p, 0, HISTORY_RSRC_SIZE, 0);
		    record.entry_size = sizeof(struct history_entry);
		    record.entry_cnt = 0;
		    DmWrite(p, 0, &record, sizeof record);
		    MemHandleUnlock(h);
		}
	    }
	    if (h != NULL) {
		if (MemHandleSize(h) == HISTORY_RSRC_SIZE
		 && (p = MemHandleLock(h)) != NULL) {
		    if (p->entry_size == sizeof(struct history_entry)
		     && p->entry_cnt >= 0
		     && p->entry_cnt <= HISTORY_MAX_COUNT) {
			for (i = 0; i < p->entry_cnt; ++i) {
			    if (p->entries[i].creator == entry.creator
			     && p->entries[i].type == entry.type) {
				break;
			    }
			}
			if (i == p->entry_cnt) {
			    if (i >= HISTORY_MAX_COUNT) {
				i = HISTORY_MAX_COUNT - 1;
			    } else {
				record = *p;
				++record.entry_cnt;
				DmWrite(p, 0, &record, sizeof record);
			    }
			}
			for ( ; i > 0; --i) {
			    DmWrite(p, sizeof(struct history_record)
				     + sizeof(struct history_entry) * i,
				    &p->entries[i - 1],
				    sizeof(struct history_entry));
			}
			DmWrite(p, sizeof(struct history_record),
				&entry, sizeof entry);
		    }
		    MemHandleUnlock(h);
		}
		DmReleaseResource(h);
	    }
	}
	DmCloseDatabase(db);
    }
}

static void switch_hook(UInt card_no, LocalID db_id)
{
    struct adal_prefs prefs;
    char name[dmDBNameLength];
    UInt attr;
    ULong type;
    ULong creator;
    DWord rom_version;

    get_prefs(&prefs);

    if (prefs.record_history == RECORD_ALL_APPS_DAS
     || prefs.record_history == RECORD_ALL_APPS) {
	if (DmDatabaseInfo(card_no, db_id, name, &attr, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator) == noErr
	 && (attr & dmHdrAttrResDB) != 0
	 && type == APP_DB_TYPE
	 && creator != LAUNCHER_CREATOR
	 && creator != LAUNCHPAD_CREATOR
	 && creator != COLAUNCHER_CREATOR
	 && creator != CLICK_LC_CREATOR
	 && creator != CLICK_LC_A_CREATOR
	 && creator != CLICK_LC_B_CREATOR
	 && creator != CLICK_LC_C_CREATOR
	 && creator != CLICK_LC_D_CREATOR
	 && (FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_version) != noErr
	  || sysGetROMVerMajor(rom_version) < 2
	  || creator != PrefGetPreference(prefLauncherAppCreator))) {
	    if (prefs.app_name_lookup) {
		get_icon_name(card_no, db_id, name);
	    }
	    record_history(type, creator, name);
	}
    }
}


/*
 * main routine
 */

Err MyHandler(UInt card_no, LocalID db_id, Word cmd, Ptr cmd_pbp)
{
    Err (*orig_trap)(UInt card_no, LocalID db_id, Word cmd, Ptr cmd_pbp);

    if (cmd == sysAppLaunchCmdNormalLaunch
     || cmd == sysAppLaunchCmdGoTo) {
	switch_hook(card_no, db_id);
    }

    if (FtrGet(CREATOR, FTR_ORIG, (DWord *) &orig_trap) == noErr) {
	return (*orig_trap)(card_no, db_id, cmd, cmd_pbp);
    }

    return noErr;
}
